import logging
import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor, as_completed

log = logging.getLogger(__name__)


class Executor():
    def __init__(self, fn, worker_num=mp.cpu_count(),
                 defaultWorkerFinished=None,
                 onFinished=None,
                 onError=None):
        self.fn = fn
        self.worker_num = worker_num
        self.onFinished = onFinished
        self.onError = onError
        self.defaultWorkerFinished = defaultWorkerFinished
        self.onWorkersFinished = dict()
        self.futures = []
        self.results = []

    def on_worker_finished(self, workerId, defaultWorkerFinished):
        self.onWorkersFinished.setdefault(workerId, defaultWorkerFinished)

    def __done_call_back(self, i):
        executor = self

        def call_back(future):
            executor.futures.remove(future)
            executor.results.append(future.result)
            onWorkerFinished = executor.onWorkersFinished.get(i)
            if onWorkerFinished:
                onWorkerFinished(i, future.result)
            elif executor.defaultWorkerFinished:
                self.defaultWorkerFinished(i, future.result)

            if len(executor.futures) == 0 and executor.onFinished:
                executor.onFinished()

        return call_back

    def start(self):
        log.info("Starting executor with worker number: " + str(self.worker_num))
        try:
            with ProcessPoolExecutor(max_workers=self.worker_num) as executor:
                for i in range(self.worker_num):
                    future = executor.submit(self.fn, i)
                    self.futures.append(future)
                    future.add_done_callback(self.__done_call_back(i))
        except Exception as ex:
            log.error(ex)
            if self.onError:
                self.onError(ex)
        return self

    def get(self):
        as_completed(self.futures)

        return self.results


class BatchExecutor():
    def __init__(self, items, worker_num=mp.cpu_count(), check_count=10000):
        self.items = items
        self.max_index = len(items)
        self.worker_num = worker_num
        self.batch_size = (self.max_index // worker_num) + 1
        self.check_count = check_count

    def run(self):
        result = []
        futures = []

        log.info("Start Batch handler Excecutor to handle data, len: " + str(self.max_index))
        try:
            with ProcessPoolExecutor(max_workers=self.worker_num) as executor:
                from_index, to_index = 0, 0
                for i in range(self.worker_num):
                    to_index = (1 + i) * self.batch_size
                    if to_index >= self.max_index:
                        to_index = self.max_index
                    futures.append(executor.submit(self.handle_items, from_index, to_index))
                    from_index = to_index
                for f in as_completed(futures):
                    result.append(f.result())
        except Exception as ex:
            log.error(ex)
        return result

    def handle_items(self, from_index, to_index):
        raise NotImplementedError()


class BatchHandlerExecutor(BatchExecutor):
    def __init__(self, items, fn, worker_num=mp.cpu_count(), check_count=10000):
        super().__init__(items, worker_num, check_count)
        self.fn = fn

    def handle_items(self, from_index, to_index):
        result = []

        for i in range(from_index, to_index):
            result.append(self.fn(i, self.items[i]))
            if i % self.check_count == 0:
                log.info("handle items %d of %d" % (i, self.max_index))
        return result



class NoneAttr:
    pass


class Inspector:
    @staticmethod
    def parse_attribute(instance, name, with_type=False, parse=False):
        parent = instance
        if parse:
            attributes = name.split('.')
            name = attributes[-1]
            for attr in attributes[:-1]:
                parent = getattr(instance, attr, NoneAttr)
                if parent == NoneAttr:
                    return None, None, None

        if not with_type:
            return parent, name, None

        field = type(parent).__dict__.get('__dataclass_fields__').get(name)
        if field is None:
            return parent, name, None
        return parent, name, field.type

    @staticmethod
    def field_type(instance, name: str, parse=False):
        _, _, type = Inspector.parse_attribute(instance, name, True, parse)
        return type

    @staticmethod
    def get_attr(instance, name, parse=False):
        parent, name, _ = Inspector.parse_attribute(instance, name, False, parse)
        if parent is None:
            return None
        return getattr(parent, name, None)

    @staticmethod
    def get_attr_with_type(instance, name, parse=False):
        parent, name, type = Inspector.parse_attribute(instance, name, True, parse)
        if parent is None:
            return None, None
        return getattr(parent, name, None), type

    @staticmethod
    def set_attr_from(fromO, toO, attr, parseFrom=True, parseTo=True):
        if fromO is None or toO is None or attr is None:
            return toO
        parent = toO
        name = attr
        t = None
        if parseTo:
            parent, name, t = Inspector.parse_attribute(parent, attr, True, parseTo)
            if parent is None or not hasattr(parent, name):
                return toO

        v = Inspector.get_attr(fromO, attr, parseFrom)

        if v is None:
            return toO

        if t is None:
            setattr(parent, name, v)
        else:
            setattr(parent, name, t(v))

        return toO

    @staticmethod
    def dict2obj(dict: dict, obj):
        for k, v in dict.items():
            setattr(obj, k, v)
        return obj

    @staticmethod
    def initialize_attr(instance, attr, parse=False):
        if instance is None or attr is None:
            return instance
        if getattr(instance, attr) is not None:
            return instance
        t = Inspector.field_type(instance, attr, parse)
        if attr is None:
            return instance
        setattr(instance, attr, t())



class AbstractProxy(object):
    """Delegates all operations (except ``.__subject__``) to another object"""
    __slots__ = ()

    def __call__(self, *args, **kw):
        return self.__subject__(*args, **kw)

    def __getattribute__(self, attr, oga=object.__getattribute__):
        subject = oga(self, '__subject__')
        if attr == '__subject__':
            return subject
        return getattr(subject, attr)

    def __setattr__(self, attr, val, osa=object.__setattr__):
        if attr == '__subject__':
            osa(self, attr, val)
        else:
            setattr(self.__subject__, attr, val)

    def __delattr__(self, attr, oda=object.__delattr__):
        if attr == '__subject__':
            oda(self, attr)
        else:
            delattr(self.__subject__, attr)

    def __bool__(self):
        return bool(self.__subject__)

    def __getitem__(self, arg):
        return self.__subject__[arg]

    def __setitem__(self, arg, val):
        self.__subject__[arg] = val

    def __delitem__(self, arg):
        del self.__subject__[arg]

    def __getslice__(self, i, j):
        return self.__subject__[i:j]

    def __setslice__(self, i, j, val):
        self.__subject__[i:j] = val

    def __delslice__(self, i, j):
        del self.__subject__[i:j]

    def __contains__(self, ob):
        return ob in self.__subject__

    for name in 'repr str hash len abs complex int long float iter'.split():
        exec("def __%s__(self): return %s(self.__subject__)" % (name, name))

    for name in 'cmp', 'coerce', 'divmod':
        exec("def __%s__(self, ob): return %s(self.__subject__, ob)" % (name, name))

    for name, op in [
        ('lt', '<'), ('gt', '>'), ('le', '<='), ('ge', '>='),
        ('eq', ' == '), ('ne', '!=')
    ]:
        exec("def __%s__(self, ob): return self.__subject__ %s ob" % (name, op))

    for name, op in [('neg', '-'), ('pos', '+'), ('invert', '~')]:
        exec("def __%s__(self): return %s self.__subject__" % (name, op))

    for name, op in [
        ('or', '|'), ('and', '&'), ('xor', '^'), ('lshift', '<<'), ('rshift', '>>'),
        ('add', '+'), ('sub', '-'), ('mul', '*'), ('div', '/'), ('mod', '%'),
        ('truediv', '/'), ('floordiv', '//')
    ]:
        exec((
                 "def __%(name)s__(self, ob):\n"
                 "    return self.__subject__ %(op)s ob\n"
                 "\n"
                 "def __r%(name)s__(self, ob):\n"
                 "    return ob %(op)s self.__subject__\n"
                 "\n"
                 "def __i%(name)s__(self, ob):\n"
                 "    self.__subject__ %(op)s=ob\n"
                 "    return self\n"
             ) % locals())

    del name, op

    # Oddball signatures
    def __index__(self):
        return self.__subject__.__index__()

    def __rdivmod__(self, ob):
        return divmod(ob, self.__subject__)

    def __pow__(self, *args):
        return pow(self.__subject__, *args)

    def __ipow__(self, ob):
        self.__subject__ **= ob
        return self

    def __rpow__(self, ob):
        return pow(ob, self.__subject__)


class ObjectProxy(AbstractProxy):
    """Proxy for a specific object"""

    __slots__ = "__subject__"

    def __init__(self, subject):
        self.__subject__ = subject


class CallbackProxy(AbstractProxy):
    """Proxy for a dynamically-chosen object"""

    __slots__ = '__callback__'

    def __init__(self, func):
        set_callback(self, func)


set_callback = CallbackProxy.__callback__.__set__
get_callback = CallbackProxy.__callback__.__get__
CallbackProxy.__subject__ = property(lambda self, gc=get_callback: gc(self)())


class LazyProxy(CallbackProxy):
    """Proxy for a lazily-obtained object, that is cached on first use"""
    __slots__ = "__cache__"


get_cache = LazyProxy.__cache__.__get__
set_cache = LazyProxy.__cache__.__set__


def __subject__(self, get_cache=get_cache, set_cache=set_cache):
    try:
        return get_cache(self)
    except AttributeError:
        set_cache(self, get_callback(self)())
        return get_cache(self)


LazyProxy.__subject__ = property(__subject__, set_cache)
del __subject__


class AbstractWrapper(AbstractProxy):
    """Mixin to allow extra behaviors and attributes on proxy instance"""
    __slots__ = ()

    def __getattribute__(self, attr, oga=object.__getattribute__):
        if attr.startswith('__'):
            subject = oga(self, '__subject__')
            if attr == '__subject__':
                return subject
            return getattr(subject, attr)
        return oga(self, attr)

    def __getattr__(self, attr, oga=object.__getattribute__):
        return getattr(oga(self, '__subject__'), attr)

    def __setattr__(self, attr, val, osa=object.__setattr__):
        if (
                attr == '__subject__' or
                hasattr(type(self), attr) and not attr.startswith('__')
        ):
            osa(self, attr, val)
        else:
            setattr(self.__subject__, attr, val)

    def __delattr__(self, attr, oda=object.__delattr__):
        if (
                attr == '__subject__' or
                hasattr(type(self), attr) and not attr.startswith('__')
        ):
            oda(self, attr)
        else:
            delattr(self.__subject__, attr)


class ObjectWrapper(ObjectProxy, AbstractWrapper):
    __slots__ = ()


class CallbackWrapper(CallbackProxy, AbstractWrapper):
    __slots__ = ()


class LazyWrapper(LazyProxy, AbstractWrapper):
    __slots__ = ()


def lazymethod(method):
    """
    Decorator for overriding methods on LazyWrapper classes, that fill in until
    lazy initialization.
    """
    import functools

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        try:
            get_cache(self)
        except AttributeError:
            return method(self, *args, **kwargs)
        else:
            if method.__name__.startswith('__'):
                return getattr(LazyProxy, method.__name__)(self, *args, **kwargs)
            else:
                return getattr(self.__subject__, method.__name__)(*args, **kwargs)

    return wrapper



import os
import keyword
from xml.sax import make_parser, handler

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
try:
    from types import StringTypes


    def is_string(x):
        return isinstance(x, StringTypes)


except ImportError:

    def is_string(x):
        return isinstance(x, str)

__version__ = "1.1.1"


class Element(object):
    """
    Representation of an XML element.
    """

    def __init__(self, name, attributes):
        self._name = name
        self._attributes = attributes
        self.children = []
        self.is_root = False
        self.cdata = ""

    def add_child(self, element):
        """
        Store child elements.
        """
        self.children.append(element)

    def add_cdata(self, cdata):
        """
        Store cdata
        """
        self.cdata = self.cdata + cdata

    def get_attribute(self, key):
        """
        Get attributes by key
        """
        return self._attributes.get(key)

    def get_elements(self, name=None):
        """
        Find a child element by name
        """
        if name:
            return [e for e in self.children if e._name == name]
        else:
            return self.children

    def __getitem__(self, key):
        return self.get_attribute(key)

    def __getattr__(self, key):
        matching_children = [x for x in self.children if x._name == key]
        if matching_children:
            if len(matching_children) == 1:
                self.__dict__[key] = matching_children[0]
                return matching_children[0]
            else:
                self.__dict__[key] = matching_children
                return matching_children
        else:
            raise AttributeError("'%s' has no attribute '%s'" % (self._name, key))

    def __hasattribute__(self, name):
        if name in self.__dict__:
            return True
        return any(x._name == name for x in self.children)

    def __iter__(self):
        yield self

    def __str__(self):
        return "Element <%s> with attributes %s, children %s and cdata %s" % (
            self._name,
            self._attributes,
            self.children,
            self.cdata,
        )

    def __repr__(self):
        return "Element(name = %s, attributes = %s, cdata = %s)" % (
            self._name,
            self._attributes,
            self.cdata,
        )

    def __nonzero__(self):
        return self.is_root or self._name is not None

    def __eq__(self, val):
        return self.cdata == val

    def __dir__(self):
        children_names = [x._name for x in self.children]
        return children_names

    def __len__(self):
        return len(self.children)

    def __contains__(self, key):
        return key in dir(self)


class Handler(handler.ContentHandler):
    """
    SAX handler which creates the Python object structure out of ``Element``s
    """

    def __init__(self):
        super().__init__()
        self.root = Element(None, None)
        self.root.is_root = True
        self.elements = []

    def startElement(self, name, attributes):
        name = name.replace("-", "_")
        name = name.replace(".", "_")
        name = name.replace(":", "_")

        # adding trailing _ for keywords
        if keyword.iskeyword(name):
            name += "_"

        attrs = dict()
        for k, v in attributes.items():
            attrs[k] = v
        element = Element(name, attrs)
        if len(self.elements) > 0:
            self.elements[-1].add_child(element)
        else:
            self.root.add_child(element)
        self.elements.append(element)

    def endElement(self, name):
        self.elements.pop()

    def characters(self, cdata):
        self.elements[-1].add_cdata(cdata)


def parse(filename, **parser_features):
    """
    Interprets the given string as a filename, URL or XML data string,
    parses it and returns a Python object which represents the given
    document.
    Extra arguments to this function are treated as feature values to pass
    to ``parser.setFeature()``. For example, ``feature_external_ges=False``
    will set ``xml.sax.handler.feature_external_ges`` to False, disabling
    the parser's inclusion of external general (text) entities such as DTDs.
    Raises ``ValueError`` if the first argument is None / empty string.
    Raises ``AttributeError`` if a requested xml.sax feature is not found in
    ``xml.sax.handler``.
    Raises ``xml.sax.SAXParseException`` if something goes wrong
    during parsing.
    """
    if filename is None or (is_string(filename) and filename.strip()) == "":
        raise ValueError("parse() takes a filename, URL or XML string")
    parser = make_parser()
    for feature, value in parser_features.items():
        parser.setFeature(getattr(handler, feature), value)
    sax_handler = Handler()
    parser.setContentHandler(sax_handler)
    if is_string(filename) and (os.path.exists(filename) or is_url(filename)):
        parser.parse(filename)
    else:
        if hasattr(filename, "read"):
            parser.parse(filename)
        else:
            parser.parse(StringIO(filename))

    return sax_handler.root


def is_url(string):
    """
    Checks if the given string starts with 'http(s)'.
    """
    try:
        return string.startswith("http://") or string.startswith("https://")
    except AttributeError:
        return False



from aiharness import harnessutils as utils
from aiharness.inspector import Inspector
from wrapt import ObjectProxy
import argparse


class ArgType(ObjectProxy):
    help = None
    type = None

    def __init__(self, v, h=None):
        super().__init__(v)
        self.help = h
        self.type = type(v)

    def set(self, v, h=None):
        super().__init__(self.type(v))
        self.help = h
        return self


class XmlConfiguration:
    def __init__(self, config):
        if config is None:
            raise ValueError("target config type can not be none.")

        if type(config) == type:
            self.config = config()

    def __set_xml2arg(self, groupObj, argXml):
        argName = argXml['name'].replace('-', '_')
        argObj: ArgType = getattr(groupObj, argName)
        if argObj is None:
            return
        argObj.set(argXml['default'], argXml['help'])

    def __set_xml2group(self, groupObj, groupXml):
        if not hasattr(groupXml, 'arg'):
            return
        if isinstance(groupXml.arg, list):
            for arg in groupXml.arg:
                self.__set_xml2arg(groupObj, arg)
        else:
            self.__set_xml2arg(groupObj, groupXml.arg)

    def __find_set_xml2group(self, config, groupXml):
        groupName = groupXml['name'].replace('-', '_')
        groupHelp = groupXml['help']
        groupObj = getattr(config, groupName)
        if groupObj is None:
            return
        setattr(groupObj, 'help', groupHelp)
        self.__set_xml2group(groupObj, groupXml)

    def load(self, xml_files: []):
        """
        Function for loading the xml_file into the configuration object.
        This function can be called multiple times to load multiple xml files.
        And the configuration value will be overrode by the following xml configuration.
        :param xml_file:
        :return: configuration object
        """
        if xml_files is None:
            return self.config
        for xml_file in xml_files:
            xml = utils.load_xml(xml_file)

            if xml is None:
                return self.config

            # if has group, set the args in the groups
            if hasattr(xml.configuration, 'group'):
                if isinstance(xml.configuration.group, list):
                    for group in xml.configuration.group:
                        self.__find_set_xml2group(self.config, group)
                else:
                    self.__find_set_xml2group(self.config, xml.configuration.group)
            ## set other args
            if hasattr(xml.configuration, 'arg'):
                self.__set_xml2group(self.config, xml.configuration)

        return self.config


class Arguments:
    def __init__(self, configObj=None, grouped=True):
        self.parser = argparse.ArgumentParser()
        self.destObj = configObj
        self.grouped = grouped
        self.groups = dict()
        self.arg_obj(configObj)

    def __get_type_action(self, argument: ArgType):
        action = 'store'
        t = type(ArgType)
        if t == bool:
            if argument.default:
                return t, 'store_false'
            else:
                return t, 'store_true'
        return t, action

    def __get_group(self, groupName, help=''):
        group = self.groups.get(groupName)
        if group is not None:
            return group
        group = self.parser.add_argument_group(groupName, help)
        self.groups.setdefault(groupName, group)
        return group

    def __prepare_arg(self, name, parser, group):
        argName = name
        if group is not None and not self.grouped:
            argName = group + '.' + argName
        if parser is None:
            if group is None:
                parser = self.parser
            else:
                parser = self.__get_group(group)
        return argName, parser

    def arg(self, name, argument: ArgType, parser=None, group=None):
        argName, parser = self.__prepare_arg(name, parser, group)

        t, action = self.__get_type_action(argument)

        parser.add_argument('--' + argName,
                            default=argument,
                            required=False,
                            action=action,
                            help=argument.help)
        return self

    def arg_obj(self, obj, groupName=None):
        if obj is None:
            return self
        parser = self.parser
        for k, v in obj.__dict__.items():
            if isinstance(v, ArgType):
                self.arg(k, v, parser, groupName)
            else:
                if groupName is not None:
                    continue
                if self.grouped:
                    groupHelp = ''
                    if hasattr(v, 'help'):
                        groupHelp = getattr(v, 'help')
                    parser = self.__get_group(k, groupHelp)
                self.arg_obj(v, k)
        return self

    def parse(self, args=None):
        return self.parseTo(self.destObj, args)

    def parseTo(self, dest, args=None):
        args, _ = self.parser.parse_known_args(args)
        if dest is None:
            return args

        for k, _ in args.__dict__.items():
            Inspector.set_attr_from(args, dest, k, False, True)

        return dest



import yaml
import logging
import logging.config
from aiharness import xml2object

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper


def load_yaml(file: str):
    try:
        with open(file, 'r') as stream:
            return yaml.load(stream=stream, Loader=Loader)
    except FileNotFoundError:
        print('File Not Found: ' + file)
        return None


conf = load_yaml('logging.yaml')
if conf is not None:
    logging.config.dictConfig(conf)


def getLogger(name: str):
    return logging.getLogger(name)


def getRootLogger():
    return logging.getLogger('root')


log = getRootLogger()


def load_xml(xml_file):
    try:
        return xml2object.parse(xml_file)
    except Exception as e:
        log.error(e)
    return None
